@@ -1,5 +1,5 @@
 # -------------------------------------------------------------------------------
-# (c) 2019-2021 Siemens AG
+# (c) 2019-2022 Siemens AG
 # All Rights Reserved.
 # Authors: thomas.graf@siemens.com, gernot.hillier@siemens.com
 #
@@ -267,6 +267,13 @@ class SW360:
         :raises SW360Error: if there is a negative HTTP response
         """
         resp = self.api_get(self.url + "resource/api/projects?type=" + project_type)
+
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:projects" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:projects"]
         return resp
 
@@ -280,8 +287,16 @@ class SW360:
         :raises SW360Error: if there is a negative HTTP response
         """
         projects = self.get_projects()
-        projects = projects["_embedded"]["sw360:projects"]
         resp = []
+
+        if "_embedded" not in projects:
+            return resp
+
+        if "sw360:projects" not in projects["_embedded"]:
+            return resp
+
+        projects = projects["_embedded"]["sw360:projects"]
+
         for key in projects:
             resp.append(key["name"] + ", " + key["version"])
 
@@ -302,6 +317,12 @@ class SW360:
         if not resp:
             return None
 
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:projects" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:projects"]
         return resp
 
@@ -324,6 +345,12 @@ class SW360:
         if not resp:
             return None
 
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:projects" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:projects"]
         return resp
 
@@ -346,6 +373,12 @@ class SW360:
         if not resp:
             return None
 
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:projects" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:projects"]
         return resp
 
@@ -360,11 +393,17 @@ class SW360:
         :rtype: list of JSON project objects
         :raises SW360Error: if there is a negative HTTP response
         """
-        full_url = self.url + "resource/api/projects?tag=" + tag
+        full_url = self.url + "resource/api/projects?tag=" + tag + "&luceneSearch=true"
         resp = self.api_get(full_url)
         if not resp:
             return None
 
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:projects" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:projects"]
         return resp
 
@@ -446,21 +485,36 @@ class SW360:
 
         raise SW360Error(response, url)
 
-    def update_project(self, project, project_id):
+    def update_project(self, project: dict, project_id: str, add_subprojects=False):
         """Update an existing project
 
         API endpoint: PATCH /projects
 
         :param project: the new project data
         :param project_id: the id of the project to be deleted
+        :param add_subprojects: optional parameter only to add new sub-projects
         :type project: JSON
         :type project_id: string
+        :type add_subprojects: bool
         :return: SW360 result
         :rtype: JSON SW360 result object
         :raises SW360Error: if there is a negative HTTP response
         """
-        # 2019-04-03: error 405 - method not allowed
+        if not project_id:
+            raise SW360Error(message="No project id provided!")
+
         url = self.url + "resource/api/projects/" + project_id
+
+        if add_subprojects:
+            current = self.get_project(project_id)
+            if (current is not None and "linkedProjects" in current):
+                for sp in current["linkedProjects"]:
+                    pid = self.get_id_from_href(sp["project"])
+                    if pid not in project["linkedProjects"]:
+                        nsp = {}
+                        nsp["projectRelationship"] = sp.get("relation", "CONTAINED")
+                        project["linkedProjects"][pid] = nsp
+
         response = requests.patch(url, json=project, headers=self.api_headers)
 
         if response.ok:
@@ -576,6 +630,78 @@ class SW360:
         resp = self.api_get(self.url + "resource/api/projects/usedBy/" + project_id)
         return resp
 
+    def duplicate_project(self, project_id: str, new_version: str):
+        """Create a copy of an exisiting project.
+
+        API endpoint: GET /projects/duplicate/{id}
+
+        :param project_id: the id of the exisiting project
+        :type project_id: string
+        :param new_version: the version of the new project
+        :type new_version: string
+        :return: the newly created project
+        :rtype: JSON object
+        :raises SW360Error: if there is a negative HTTP response
+        """
+
+        if not project_id:
+            raise SW360Error(message="No project id provided!")
+
+        project_details = {}
+        project_details["version"] = new_version
+        # force clearing state to OPEN
+        project_details["clearingState"] = "OPEN"
+
+        url = self.url + "resource/api/projects/duplicate/" + project_id
+        response = requests.post(
+            url, json=project_details, headers=self.api_headers
+        )
+
+        if response.ok:
+            return response.json()
+
+        raise SW360Error(response, url)
+
+    def update_project_release_relationship(
+        self, project_id: str, release_id: str, new_state: str,
+            new_relation: str, comment: str):
+        """Update the relationship for a specific release of a project
+
+        API endpoint PATCH /projects/{pid}/release{rid}
+
+        :param project_id: the id of the exisiting project
+        :type project_id: string
+        :param release_id: the id of the release to be requested
+        :type release_id: string
+        :param new_state: the new mainline state of the release, one of
+         (OPEN, MAINLINE, SPECIFIC, PHASEOUT, DENIED)
+        :type new_state: string
+        :param new_relation: the new relation of the release, one of
+         (CONTAINED, REFERRED, UNKNOWN, DYNAMICALLY_LINKED, STATICALLY_LINKED, SIDE_BY_SIDE,
+         STANDALONE, INTERNAL_USE, OPTIONAL, TO_BE_REPLACED, CODE_SNIPPET)
+        :type new_relation: string
+        :param comment: a comment
+        :type comment: string
+        """
+        if not project_id:
+            raise SW360Error(message="No project id provided!")
+
+        if not release_id:
+            raise SW360Error(message="No release id provided!")
+
+        relation = {}
+        relation["releaseRelation"] = new_relation
+        relation["mainlineState"] = new_state
+        relation["comment"] = comment
+
+        url = self.url + "resource/api/projects/" + project_id + "/release/" + release_id
+        response = requests.patch(url, json=relation, headers=self.api_headers)
+
+        if response.ok:
+            return response.json()
+
+        raise SW360Error(response, url)
+
     # ----- Releases ---------------------------------------------------------
 
     def get_release(self, release_id):
@@ -820,6 +946,12 @@ class SW360:
         if not resp:
             return None
 
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:components" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:components"]
         return resp
 
@@ -840,6 +972,12 @@ class SW360:
         if not resp:
             return None
 
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:components" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:components"]
         return resp
 
@@ -1063,6 +1201,12 @@ class SW360:
         if not resp:
             return None
 
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:vendors" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:vendors"]
         return resp
 
@@ -1180,6 +1324,13 @@ class SW360:
             + resource_id
             + "/attachments"
         )
+
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:attachments" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:attachments"]
         return resp
 
@@ -1415,6 +1566,12 @@ class SW360:
         if not resp:
             return None
 
+        if "_embedded" not in resp:
+            return None
+
+        if "sw360:licenses" not in resp["_embedded"]:
+            return None
+
         resp = resp["_embedded"]["sw360:licenses"]
         return resp
 
